@@ -23,7 +23,6 @@
 #include "ksycocautils_p.h"
 #include "ksycocatype.h"
 #include "ksycocafactory_p.h"
-#include "kmemfile_p.h"
 #include "kconfiggroup.h"
 #include "ksharedconfig.h"
 #include "sycocadebug.h"
@@ -33,33 +32,35 @@
 #include <QtCore/QCoreApplication>
 #include <QtCore/QFile>
 #include <QtCore/QFileInfo>
-#include <QtCore/QBuffer>
-#include <QProcess>
-#include <QDBusConnection>
-#include <QDBusConnectionInterface>
-#include <QDBusInterface>
-#include <QDBusReply>
+#include <QThread>
+#include <QThreadStorage>
+#include <QMetaMethod>
 
 #include <stdlib.h>
 #include <fcntl.h>
+#include <QDir>
+#include <kmimetypefactory_p.h>
+#include <kservicetypefactory_p.h>
+#include <kservicegroupfactory_p.h>
+#include <kservicefactory_p.h>
+#include <QCryptographicHash>
 
+#include "kbuildsycoca_p.h"
 #include "ksycocadevices_p.h"
 
-// TODO: remove mmap() from kdewin32 and use QFile::mmap() when needed
-#ifdef Q_OS_WIN
-#undef HAVE_MMAP
+#ifdef Q_OS_UNIX
+#include <utime.h>
+#include <sys/time.h>
 #endif
+
 /**
  * Sycoca file version number.
  * If the existing file is outdated, it will not get read
- * but instead we'll ask kded to regenerate a new one...
+ * but instead we'll regenerate a new one.
+ * However running apps should still be able to read it, so
+ * only add to the data, never remove/modify.
  */
-#define KSYCOCA_VERSION 300
-
-/**
- * Sycoca file name, used internally (by kbuildsycoca)
- */
-#define KSYCOCA_FILENAME "ksycoca5"
+#define KSYCOCA_VERSION 303
 
 #if HAVE_MADVISE || HAVE_MMAP
 #include <sys/mman.h> // This #include was checked when looking for posix_madvise
@@ -69,7 +70,10 @@
 #define MAP_FAILED ((void *) -1)
 #endif
 
-static bool s_autoRebuild = true;
+QDataStream &operator>>(QDataStream &in, KSycocaHeader &h) {
+    in >> h.prefixes >> h.timeStamp >> h.language >> h.updateSignature;
+    return in;
+}
 
 // The following limitations are in place:
 // Maximum length of a single string: 8192 bytes
@@ -81,16 +85,23 @@ static bool s_autoRebuild = true;
 
 Q_DECLARE_OPERATORS_FOR_FLAGS(KSycocaPrivate::BehaviorsIfNotFound)
 
-KSycocaPrivate::KSycocaPrivate()
+KSycocaPrivate::KSycocaPrivate(KSycoca *q)
     : databaseStatus(DatabaseNotOpen),
       readError(false),
       timeStamp(0),
       m_databasePath(),
       updateSig(0),
+      m_haveListeners(false),
+      m_globalDatabase(false),
+      q(q),
       sycoca_size(0),
       sycoca_mmap(0),
       m_mmapFile(0),
-      m_device(0)
+      m_device(0),
+      m_mimeTypeFactory(0),
+      m_serviceTypeFactory(0),
+      m_serviceFactory(0),
+      m_serviceGroupFactory(0)
 {
 #ifdef Q_OS_WIN
     /*
@@ -114,7 +125,7 @@ void KSycocaPrivate::setStrategyFromString(const QString &strategy)
     } else if (strategy == QLatin1String("sharedmem")) {
         m_sycocaStrategy = StrategyMemFile;
     } else if (!strategy.isEmpty()) {
-        qWarning() << "Unknown sycoca strategy:" << strategy;
+        qCWarning(SYCOCA) << "Unknown sycoca strategy:" << strategy;
     }
 }
 
@@ -130,23 +141,25 @@ bool KSycocaPrivate::tryMmap()
     }
     fcntl(m_mmapFile->handle(), F_SETFD, FD_CLOEXEC);
     sycoca_size = m_mmapFile->size();
-    sycoca_mmap = (const char *) mmap(0, sycoca_size,
-                                      PROT_READ, MAP_SHARED,
-                                      m_mmapFile->handle(), 0);
+    void *mmapRet = mmap(0, sycoca_size,
+                       PROT_READ, MAP_SHARED,
+                       m_mmapFile->handle(), 0);
     /* POSIX mandates only MAP_FAILED, but we are paranoid so check for
        null pointer too.  */
-    if (sycoca_mmap == (const char *) MAP_FAILED || sycoca_mmap == 0) {
+    if (mmapRet == MAP_FAILED || mmapRet == 0) {
         qCDebug(SYCOCA).nospace() << "mmap failed. (length = " << sycoca_size << ")";
         sycoca_mmap = 0;
         return false;
     } else {
+        sycoca_mmap = static_cast<const char *>(mmapRet);
 #if HAVE_MADVISE
-        (void) posix_madvise((void *)sycoca_mmap, sycoca_size, POSIX_MADV_WILLNEED);
+        (void) posix_madvise(mmapRet, sycoca_size, POSIX_MADV_WILLNEED);
 #endif // HAVE_MADVISE
         return true;
     }
-#endif // HAVE_MMAP
+#else
     return false;
+#endif // HAVE_MMAP
 }
 
 int KSycoca::version()
@@ -182,14 +195,39 @@ private:
 
 Q_GLOBAL_STATIC(KSycocaSingleton, ksycocaInstance)
 
-// Read-only constructor
-KSycoca::KSycoca()
-    : d(new KSycocaPrivate)
+QString KSycocaPrivate::findDatabase()
 {
-    QDBusConnection::sessionBus().connect(QString(), QString(),
-                                          QString::fromLatin1("org.kde.KSycoca"),
-                                          QString::fromLatin1("notifyDatabaseChanged"),
-                                          this, SLOT(notifyDatabaseChanged(QStringList)));
+    Q_ASSERT(databaseStatus == DatabaseNotOpen);
+
+    m_globalDatabase = false;
+    QString path = KSycoca::absoluteFilePath();
+    QFileInfo info(path);
+    bool canRead = info.isReadable();
+    if (!canRead) {
+        path = KSycoca::absoluteFilePath(KSycoca::GlobalDatabase);
+        if (!path.isEmpty()) {
+            info.setFile(path);
+            canRead = info.isReadable();
+            if (canRead) {
+                m_globalDatabase = true;
+            }
+        }
+    }
+    if (canRead) {
+        if (m_haveListeners) {
+            m_fileWatcher.addFile(path);
+        }
+        return path;
+    }
+    return QString();
+}
+
+// Read-only constructor
+// One instance per thread
+KSycoca::KSycoca()
+    : d(new KSycocaPrivate(this))
+{
+    connect(&d->m_fileWatcher, &KDirWatch::created, this, [this](){ d->slotDatabaseChanged(); });
 }
 
 bool KSycocaPrivate::openDatabase(bool openDummyIfNotFound)
@@ -197,26 +235,19 @@ bool KSycocaPrivate::openDatabase(bool openDummyIfNotFound)
     Q_ASSERT(databaseStatus == DatabaseNotOpen);
 
     delete m_device; m_device = 0;
-    QString path = KSycoca::absoluteFilePath();
 
-    bool canRead = QFileInfo(path).isReadable();
-    qCDebug(SYCOCA) << "Trying to open ksycoca from" << path;
-    if (!canRead) {
-        path = KSycoca::absoluteFilePath(KSycoca::GlobalDatabase);
-        if (!path.isEmpty()) {
-            qCDebug(SYCOCA) << "Trying to open global ksycoca from " << path;
-            canRead = QFileInfo(path).isReadable();
-        }
+    if (m_databasePath.isEmpty()) {
+        m_databasePath = findDatabase();
     }
 
     bool result = true;
-    if (canRead) {
-        m_databasePath = path;
+    if (!m_databasePath.isEmpty()) {
+        qCDebug(SYCOCA) << "Opening ksycoca from" << m_databasePath;
+        m_dbLastModified = QFileInfo(m_databasePath).lastModified();
         checkVersion();
     } else { // No database file
         //qCDebug(SYCOCA) << "Could not open ksycoca";
         m_databasePath.clear();
-        databaseStatus = NoDatabase;
         if (openDummyIfNotFound) {
             // We open a dummy database instead.
             //qCDebug(SYCOCA) << "No database, opening a dummy one.";
@@ -265,7 +296,7 @@ KSycocaAbstractDevice *KSycocaPrivate::device()
         if (!device) {
             device = new KSycocaFileDevice(m_databasePath);
             if (!device->device()->open(QIODevice::ReadOnly)) {
-                qWarning() << "Couldn't open" << m_databasePath << "even though it is readable? Impossible.";
+                qCWarning(SYCOCA) << "Couldn't open" << m_databasePath << "even though it is readable? Impossible.";
                 //delete device; device = 0; // this would crash in the return statement...
             }
         }
@@ -289,12 +320,67 @@ QDataStream *&KSycocaPrivate::stream()
     return m_device->stream();
 }
 
+void KSycocaPrivate::slotDatabaseChanged()
+{
+    // We don't have information anymore on what resources changed, so emit them all
+    changeList = QStringList() << QStringLiteral("services") << QStringLiteral("servicetypes") << QStringLiteral("xdgdata-mime") << QStringLiteral("apps");
+
+    qCDebug(SYCOCA) << QThread::currentThread() << "got a notifyDatabaseChanged signal";
+    // kbuildsycoca tells us the database file changed
+    // We would have found out in the next call to ensureCacheValid(), but for
+    // now keep the call to closeDatabase, to help refcounting to 0 the old mmaped file earlier.
+    closeDatabase();
+    // Start monitoring the new file right away
+    m_databasePath = findDatabase();
+
+    // Now notify applications
+    emit q->databaseChanged();
+    emit q->databaseChanged(changeList);
+}
+
+KMimeTypeFactory *KSycocaPrivate::mimeTypeFactory()
+{
+    if (!m_mimeTypeFactory) {
+        m_mimeTypeFactory = new KMimeTypeFactory(q);
+    }
+    return m_mimeTypeFactory;
+}
+
+KServiceTypeFactory *KSycocaPrivate::serviceTypeFactory()
+{
+    if (!m_serviceTypeFactory) {
+        m_serviceTypeFactory = new KServiceTypeFactory(q);
+    }
+    return m_serviceTypeFactory;
+}
+
+KServiceFactory *KSycocaPrivate::serviceFactory()
+{
+    if (!m_serviceFactory) {
+        m_serviceFactory = new KServiceFactory(q);
+    }
+    return m_serviceFactory;
+}
+
+KServiceGroupFactory *KSycocaPrivate::serviceGroupFactory()
+{
+    if (!m_serviceGroupFactory) {
+        m_serviceGroupFactory = new KServiceGroupFactory(q);
+    }
+    return m_serviceGroupFactory;
+}
+
+// Add local paths to the list of dirs we got from the global database
+void KSycocaPrivate::addLocalResourceDir(const QString &path)
+{
+    // If any local path is more recent than the time the global sycoca was created, build a local sycoca.
+    allResourceDirs.insert(path, timeStamp);
+}
+
 // Read-write constructor - only for KBuildSycoca
 KSycoca::KSycoca(bool /* dummy */)
-    : d(new KSycocaPrivate)
+    : d(new KSycocaPrivate(this))
 {
-    // This instance was not created by the singleton, but by a direct call to new!
-    ksycocaInstance()->setSycoca(this);
 }
 
 KSycoca *KSycoca::self()
@@ -313,7 +399,7 @@ KSycoca::~KSycoca()
     //    ksycocaInstance->self = 0;
 }
 
-bool KSycoca::isAvailable()
+bool KSycoca::isAvailable() // TODO KF6: make it non-static (mostly useful for unittests)
 {
     return self()->d->checkDatabase(KSycocaPrivate::IfNotFoundDoNothing/* don't open dummy db if not found */);
 }
@@ -329,10 +415,14 @@ void KSycocaPrivate::closeDatabase()
     // refcounted, and deleted when the last thread is done with them
     qDeleteAll(m_factories);
     m_factories.clear();
+
+    m_mimeTypeFactory = 0;
+    m_serviceFactory = 0;
+    m_serviceTypeFactory = 0;
+    m_serviceGroupFactory = 0;
+
 #if HAVE_MMAP
     if (sycoca_mmap) {
-        //QBuffer *buf = static_cast<QBuffer*>(device);
-        //buf->buffer().clear();
         // Solaris has munmap(char*, size_t) and everything else should
         // be happy with a char* for munmap(void*, size_t)
         munmap(const_cast<char *>(sycoca_mmap), sycoca_size);
@@ -342,6 +432,7 @@ void KSycocaPrivate::closeDatabase()
 #endif
 
     databaseStatus = DatabaseNotOpen;
+    m_databasePath.clear();
     timeStamp = 0;
 }
 
@@ -357,23 +448,6 @@ bool KSycoca::isChanged(const char *type)
 }
 #endif
 
-void KSycoca::notifyDatabaseChanged(const QStringList &changeList)
-{
-    d->changeList = changeList;
-    //qCDebug(SYCOCA) << QThread::currentThread() << "got a notifyDatabaseChanged signal" << changeList;
-    // kbuildsycoca tells us the database file changed
-    // Close the database and forget all about what we knew
-    // The next call to any public method will recreate
-    // everything that's needed.
-    d->closeDatabase();
-
-    // Now notify applications
-#ifndef KSERVICE_NO_DEPRECATED
-    emit databaseChanged();
-#endif
-    emit databaseChanged(changeList);
-}
-
 QDataStream *KSycoca::findEntry(int offset, KSycocaType &type)
 {
     QDataStream *str = stream();
@@ -410,7 +484,8 @@ bool KSycocaPrivate::checkVersion()
     }
 }
 
-// Allow unittests to skip the kded requirement
+// This is now completely useless. KF6: remove
+extern KSERVICE_EXPORT bool kservice_require_kded;
 KSERVICE_EXPORT bool kservice_require_kded = true;
 
 // If it returns true, we have a valid database and the stream has rewinded to the beginning
@@ -425,82 +500,26 @@ bool KSycocaPrivate::checkDatabase(BehaviorsIfNotFound ifNotFound)
 
     closeDatabase(); // close the dummy one
 
-    // We can only use the installed ksycoca file if kded is running,
-    // since kded is what keeps the file uptodate.
-    QDBusConnectionInterface *bus = QDBusConnection::sessionBus().interface();
-    const bool kdedRunning = bus->isServiceRegistered(QString::fromLatin1("org.kde.kded5")) ||
-                             qAppName() == "kbuildsycoca5";
-
     // Check if new database already available
-    if ((kdedRunning || !kservice_require_kded) && openDatabase(ifNotFound & IfNotFoundOpenDummy)) {
+    if (openDatabase(ifNotFound & IfNotFoundOpenDummy)) {
         if (checkVersion()) {
-            // Database exists, and version is ok.
+            // Database exists, and version is ok, we can read it.
+
+            if (qAppName() != KBUILDSYCOCA_EXENAME && ifNotFound != IfNotFoundDoNothing) {
+
+                // Ensure it's uptodate, rebuild if needed
+                checkDirectories();
+
+                // Don't check again for some time
+                m_lastCheck.start();
+            }
+
             return true;
         }
     }
 
     if (ifNotFound & IfNotFoundRecreate) {
-        // Ask kded to rebuild ksycoca
-        // (so that it's not only built, but also kept up-to-date...)
-        bool kdedRunning = false;
-        QDBusInterface sycoca(QLatin1String("org.kde.kded5"), QLatin1String("/kbuildsycoca"));
-        if (!bus->isServiceRegistered(QLatin1String("org.kde.kded5"))) {
-            // kded isn't even running: start it
-            QDBusReply<void> reply = bus->startService(QLatin1String("org.kde.kded5"));
-            if (!reply.isValid()) {
-                // kded isn't even available, fall back to kbuildsycoca
-                qWarning() << "Couldn't start kded5 from org.kde.kded5.service:" << reply.error() << ", falling back to running" << KBUILDSYCOCA_EXENAME;
-                QProcess proc;
-                const QString kbuildsycoca = QStandardPaths::findExecutable(KBUILDSYCOCA_EXENAME);
-                if (!kbuildsycoca.isEmpty()) {
-                    QStringList args;
-                    if (QStandardPaths::isTestModeEnabled()) {
-                        args << "--testmode";
-                    }
-                    proc.setProcessChannelMode(QProcess::MergedChannels); // silence kbuildsycoca output
-                    proc.start(kbuildsycoca, args);
-                    proc.waitForFinished();
-                }
-            } else {
-                //qCDebug(SYCOCA) << "kded5 registered";
-                kdedRunning = true;
-                if (QStandardPaths::isTestModeEnabled()) {
-                    sycoca.call(QLatin1String("enableTestMode"));
-                }
-            }
-        } else {
-            //qCDebug(SYCOCA) << "kded5 found";
-            if (QStandardPaths::isTestModeEnabled()) {
-                const QDBusReply<bool> testMode = sycoca.call(QLatin1String("isTestModeEnabled"));
-                if (!testMode.value()) {
-                    qWarning() << "This unit test uses ksycoca, it needs to be run in a separate DBus session, so that kded can be started in 'test mode'.";
-                    qWarning() << "KSycoca updates will very likely fail unless you do that.";
-                    qWarning() << "`eval dbus-launch` ; make test";
-                    // Idea for the future: move kbuildsycoca stuff to its own kded module, and use
-                    // the same module with a different name, for test mode.
-                    // On the other hand, the use of other kded modules (cookies, timezone, etc.)
-                    // is also better separated from the user's kded anyway.
-                }
-            }
-        }
-
-        //qCDebug(SYCOCA) << "We have no database.... asking kded to create it";
-        if (kdedRunning) {
-            sycoca.call(QLatin1String("recreate"));
-        }
-
-        closeDatabase(); // close the dummy one
-
-        // Ok, the new database should be here now, open it.
-        if (!openDatabase(ifNotFound & IfNotFoundOpenDummy)) {
-            qCDebug(SYCOCA) << "Still no database...";
-            return false; // Still no database - uh oh
-        }
-        if (!checkVersion()) {
-            qCDebug(SYCOCA) << "Still outdated...";
-            return false; // Still outdated - uh oh
-        }
-        return true;
+        return buildSycoca();
     }
 
     return false;
@@ -521,7 +540,7 @@ QDataStream *KSycoca::findFactory(KSycocaFactoryId id)
     while (true) {
         *str >> aId;
         if (aId == 0) {
-            qWarning() << "Error, KSycocaFactory (id =" << int(id) << ") not found!";
+            qCWarning(SYCOCA) << "Error, KSycocaFactory (id =" << int(id) << ") not found!";
             break;
         }
         *str >> aOffset;
@@ -534,13 +553,21 @@ QDataStream *KSycoca::findFactory(KSycocaFactoryId id)
     return 0;
 }
 
-QString KSycoca::kfsstnd_prefixes()
+bool KSycoca::needsRebuild()
 {
+    return d->needsRebuild();
+}
+
+KSycocaHeader KSycocaPrivate::readSycocaHeader()
+{
+    KSycocaHeader header;
     // do not try to launch kbuildsycoca from here; this code is also called by kbuildsycoca.
-    if (!d->checkDatabase(KSycocaPrivate::IfNotFoundDoNothing)) {
-        return QString();
+    if (!checkDatabase(KSycocaPrivate::IfNotFoundDoNothing)) {
+        return header;
     }
     QDataStream *str = stream();
+    qint64 oldPos = str->device()->pos();
+
     Q_ASSERT(str);
     qint32 aId;
     qint32 aOffset;
@@ -554,44 +581,163 @@ QString KSycoca::kfsstnd_prefixes()
         }
     }
     // We now point to the header
-    QString prefixes;
-    KSycocaUtilsPrivate::read(*str, prefixes);
-    *str >> d->timeStamp;
-    KSycocaUtilsPrivate::read(*str, d->language);
-    *str >> d->updateSig;
-    KSycocaUtilsPrivate::read(*str, d->allResourceDirs);
-    return prefixes;
+    QStringList directoryList;
+    *str >> header >> directoryList;
+    allResourceDirs.clear();
+    for (int i = 0; i < directoryList.count(); ++i) {
+        qint64 mtime;
+        *str >> mtime;
+        allResourceDirs.insert(directoryList.at(i), mtime);
+    }
+
+    str->device()->seek(oldPos);
+
+    timeStamp = header.timeStamp;
+
+    // for the useless public accessors. KF6: remove these two lines, the accessors and the vars.
+    language = header.language;
+    updateSig = header.updateSignature;
+
+    if (m_globalDatabase) {
+        // The global database doesn't point to the user's local dirs, but we need to check them too
+        // to react on something being created there
+        const QString dataHome = QStandardPaths::writableLocation(QStandardPaths::GenericDataLocation);
+        addLocalResourceDir(dataHome + QLatin1String("/kservices5"));
+        addLocalResourceDir(dataHome + QLatin1String("/kservicetypes5"));
+        addLocalResourceDir(dataHome + QLatin1String("/mime"));
+        addLocalResourceDir(dataHome + QLatin1String("/applications"));
+    }
+
+    return header;
+}
+
+
+class TimestampChecker
+{
+public:
+    TimestampChecker()
+        : m_now(QDateTime::currentDateTime())
+    {}
+
+    // Check times of last modification of all directories on which ksycoca depends,
+    // If none of them is newer than the mtime we stored for that directory at the
+    // last rebuild, this means that there's no need to rebuild ksycoca.
+    bool checkTimestamps(const QMap<QString, qint64> &dirs)
+    {
+        Q_ASSERT(!dirs.isEmpty());
+        qCDebug(SYCOCA) << "checking file timestamps";
+        for (auto it = dirs.begin(); it != dirs.end(); ++it) {
+            const QString dir = it.key();
+            const qint64 lastStamp = it.value();
+
+            auto visitor = [&] (const QFileInfo &fi) {
+                const QDateTime mtime = fi.lastModified();
+                if (mtime.toMSecsSinceEpoch() > lastStamp) {
+                    if (mtime > m_now) {
+                        qCDebug(SYCOCA) << fi.filePath() << "has a modification time in the future" << mtime;
+                    }
+                    qCDebug(SYCOCA) << "timestamp changed:" << fi.filePath() << mtime << ">" << QDateTime::fromMSecsSinceEpoch(lastStamp);
+                    // no need to continue search
+                    return false;
+                }
+
+                return true;
+            };
+
+            if (!KSycocaUtilsPrivate::visitResourceDirectory(dir, visitor)) {
+                return false;
+            }
+        }
+        return true;
+    }
+
+private:
+    QDateTime m_now;
+};
+
+void KSycocaPrivate::checkDirectories()
+{
+    if (needsRebuild()) {
+        buildSycoca();
+    }
+}
+
+bool KSycocaPrivate::needsRebuild()
+{
+    if (!timeStamp && databaseStatus == DatabaseOK) {
+        (void) readSycocaHeader();
+    }
+    // these days timeStamp is really a "bool headerFound", the value itself doesn't matter...
+    // KF6: replace it with bool.
+    return timeStamp != 0 && !TimestampChecker().checkTimestamps(allResourceDirs);
+}
+
+bool KSycocaPrivate::buildSycoca()
+{
+    KBuildSycoca builder;
+    if (!builder.recreate()) {
+        return false; // error
+    }
+
+    closeDatabase(); // close the dummy one
+
+    // Ok, the new database should be here now, open it.
+    if (!openDatabase()) {
+        qCDebug(SYCOCA) << "Still no database...";
+        return false; // Still no database - uh oh
+    }
+    if (!checkVersion()) {
+        qCDebug(SYCOCA) << "Still outdated...";
+        return false; // Still outdated - uh oh
+    }
+    return true;
 }
 
 quint32 KSycoca::timeStamp()
 {
     if (!d->timeStamp) {
-        (void) kfsstnd_prefixes();
+        (void) d->readSycocaHeader();
     }
-    return d->timeStamp;
+    return d->timeStamp / 1000; // from ms to s
 }
 
 quint32 KSycoca::updateSignature()
 {
     if (!d->timeStamp) {
-        (void) kfsstnd_prefixes();
+        (void) d->readSycocaHeader();
     }
     return d->updateSig;
 }
 
 QString KSycoca::absoluteFilePath(DatabaseType type)
 {
+    const QStringList paths = QStandardPaths::standardLocations(QStandardPaths::GenericDataLocation);
+    QString suffix = QLatin1Char('_') + QLocale().bcp47Name();
+
     if (type == GlobalDatabase) {
-        QString path = QStandardPaths::locate(QStandardPaths::GenericDataLocation, QString::fromLatin1("kservices5/" KSYCOCA_FILENAME));
+        const QString fileName = QStringLiteral("ksycoca5") + suffix;
+        QString path = QStandardPaths::locate(QStandardPaths::GenericDataLocation, QStringLiteral("ksycoca/") + fileName);
         if (path.isEmpty()) {
-            return QStandardPaths::writableLocation(QStandardPaths::GenericDataLocation) + QString::fromLatin1("/kservices5/" KSYCOCA_FILENAME);
+            // No existing global DB found, maybe we are going to make a new one.
+            // We don't want it in $HOME, so skip the first entry in <paths> and pick the second one.
+            if (paths.count() == 1) { // unlikely
+                return QString();
+            }
+            return paths.at(1) + QStringLiteral("/ksycoca/") + fileName;
         }
         return path;
     }
 
     const QByteArray ksycoca_env = qgetenv("KDESYCOCA");
     if (ksycoca_env.isEmpty()) {
-        return QStandardPaths::writableLocation(QStandardPaths::GenericCacheLocation) + QLatin1Char('/') + QString::fromLatin1(KSYCOCA_FILENAME);
+        const QByteArray pathHash = QCryptographicHash::hash(paths.join(QString(QLatin1Char(':'))).toUtf8(), QCryptographicHash::Sha1);
+        suffix += QLatin1Char('_') + QString::fromLatin1(pathHash.toBase64());
+        suffix.replace('/', '_');
+#ifdef Q_OS_WIN
+        suffix.replace(':', '_');
+#endif
+        const QString fileName = QStringLiteral("ksycoca5") + suffix;
+        return QStandardPaths::writableLocation(QStandardPaths::GenericCacheLocation) + QLatin1Char('/') + fileName;
     } else {
         return QFile::decodeName(ksycoca_env);
     }
@@ -600,7 +746,7 @@ QString KSycoca::absoluteFilePath(DatabaseType type)
 QString KSycoca::language()
 {
     if (d->language.isEmpty()) {
-        (void) kfsstnd_prefixes();
+        (void) d->readSycocaHeader();
     }
     return d->language;
 }
@@ -608,27 +754,23 @@ QString KSycoca::language()
 QStringList KSycoca::allResourceDirs()
 {
     if (!d->timeStamp) {
-        (void) kfsstnd_prefixes();
+        (void) d->readSycocaHeader();
     }
-    return d->allResourceDirs;
+    return d->allResourceDirs.keys();
 }
 
 void KSycoca::flagError()
 {
-    qWarning() << "ERROR: KSycoca database corruption!";
-    KSycocaPrivate *d = ksycocaInstance()->sycoca()->d;
-    if (d->readError) {
+    qCWarning(SYCOCA) << "ERROR: KSycoca database corruption!";
+    KSycoca *sycoca = self();
+    if (sycoca->d->readError) {
         return;
     }
-    d->readError = true;
-    if (s_autoRebuild) {
+    sycoca->d->readError = true;
+    if (qAppName() != KBUILDSYCOCA_EXENAME && !sycoca->isBuilding()) {
         // Rebuild the damned thing.
-        if (QProcess::execute(QStandardPaths::findExecutable(QString::fromLatin1(KBUILDSYCOCA_EXENAME))) != 0) {
-            qWarning("ERROR: Running %s failed", KBUILDSYCOCA_EXENAME);
-        }
-        // Old comment, maybe not true anymore:
-        // Do not wait until the DBUS signal from kbuildsycoca here.
-        // It deletes m_str which is a problem when flagError is called during the KSycocaFactory ctor...
+        KBuildSycoca builder;
+        (void)builder.recreate();
     }
 }
 
@@ -639,7 +781,7 @@ bool KSycoca::isBuilding()
 
 void KSycoca::disableAutoRebuild()
 {
-    s_autoRebuild = false;
+    qCWarning(SYCOCA) << "KSycoca::disableAutoRebuild() is internal, do not call it.";
 }
 
 QDataStream *&KSycoca::stream()
@@ -647,9 +789,54 @@ QDataStream *&KSycoca::stream()
     return d->stream();
 }
 
+void KSycoca::connectNotify(const QMetaMethod &signal)
+{
+    if (signal.name() == "databaseChanged" && !d->m_haveListeners) {
+        d->m_haveListeners = true;
+        if (d->m_databasePath.isEmpty()) {
+            d->m_databasePath = d->findDatabase();
+        } else {
+            d->m_fileWatcher.addFile(d->m_databasePath);
+        }
+    }
+}
+
 void KSycoca::clearCaches()
 {
     if (ksycocaInstance.exists() && ksycocaInstance()->hasSycoca())
         ksycocaInstance()->sycoca()->d->closeDatabase();
 }
 
+extern KSERVICE_EXPORT int ksycoca_ms_between_checks;
+KSERVICE_EXPORT int ksycoca_ms_between_checks = 1500;
+
+void KSycoca::ensureCacheValid()
+{
+    if (qAppName() == KBUILDSYCOCA_EXENAME) {
+        return;
+    }
+
+    if (d->databaseStatus != KSycocaPrivate::DatabaseOK) {
+        if (!d->checkDatabase(KSycocaPrivate::IfNotFoundRecreate)) {
+            return;
+        }
+    }
+
+    if (d->m_lastCheck.isValid() && d->m_lastCheck.elapsed() < ksycoca_ms_between_checks) {
+        return;
+    }
+    d->m_lastCheck.start();
+
+    // Check if the file on disk was modified since we last checked it.
+    QFileInfo info(d->m_databasePath);
+    if (info.lastModified() == d->m_dbLastModified) {
+        // Check if the watched directories were modified, then the cache needs a rebuild.
+        d->checkDirectories();
+        return;
+    }
+
+    // Close the database and forget all about what we knew.
+    // The next call to any public method will recreate
+    // everything that's needed.
+    d->closeDatabase();
+}
